% Minimal example for running the sgd+probabilistic line search on noisy
% functions. 
%
% The line search has access to exact values of sigmaf and sigmadf
%
% To get an intuition you can change the following settings
%   testfunction -- choose among 3 test functions (1, 2, 3)
%   T -- # function evaluations
%   sigmaf -- absolute noise of function values
%   sigmadf -- absolute noise of gradient evaluations 
%   verbosity -- 0: silence, 1: speak, 2: plot simple, 3: plot full (slow 
%     but good for intuition)
% and run several times. 
% 
% In this synthetic setting the noise sigmaf, sigmadf are absolute. 
% This means that the signal-to-noise ratio gets worse with dropping 
% function and gradient value (mostly closer to the minimum). 
% This also means diffusion towards the end. This is clearly visible 
% in the optimization path. Noise in real world applicatons might depend 
% on the location x, thus they need to be estimated locally.
%
% (C) 2015, Maren Mahsereci (mmahsereci@tue.mpg.de)

clear;

% == CHANGE STUFF BELOW HERE ==============================================
verbosity    = 1; % 0: silence, 1: speak, 2: plot simple, 3: plot full (slow)
ff           = @noisyFunction;
T            = 50; % max number of function evaluations
testfunction = 3;  % choose among 3 test functions (1, 2, 3)

% sythetic noise standard deviations for function value and gradient
sigmaf  = 1;
sigmadf = 1*ones(2, 1);
% == CHANGE STUFF ABOVE HERE ==============================================

switch testfunction

    case 1 % Booth2
        x1min = -10; x1max = 10; % x-limits for plotting
        x2min = -10; x2max = 10; % y-limits for plotting
        F        = @booth2;      % function handle
        x_min{1} = [1, 3];       % minimizer
        paras    = [];           % function does not have parameters
        x0       = [4; -5];      % startvalue

    case 2 % McCormick2
        x1min = -1.5; x1max = 4;       % x-limits for plotting
        x2min = -3;   x2max = 4;       % y-limits for plotting
        F        = @mccormick2;        % function handle
        x_min{1} = [-0.5472, -1.5472]; % minimizer 1
        x_min{2} = [2.5944, 1.5944];   % minimizer 2       
        paras    = [];                 % function does not have parameters
        x0       = [3; -2];            % startvalue

    case 3 % Branin2
        x1min = -5; x1max = 10;      % x-limits for plotting
        x2min = 0; x2max = 15;       % y-limits for plotting
        F        = @branin2;         % function handle
        x_min{1} = [pi, 2.275];      % minimizer 1
        x_min{2} = [-pi,12.275];     % minimizer 2
        x_min{3} = [9.42478, 2.475]; % minimizer 3
        paras    = [];               % function does not have parameters
        x0       = [5; 13];          % startvalue

end % switch

% set up function 
ff(F, sigmaf, sigmadf);

% initial function evaluation
[f, df, var_f, var_df] = ff(x0);   % noisy

search_direction = - df;   % search direction (not normalized)
xt               = x0;     % current position
alpha0           = 0.01;   % initial step size
outs.counter     = 1; % counts # function evaluations 

% storage only for plot
path            = x0;
function_values = f;

while outs.counter < T

    % -- line search ------------------------------------------------------
    [outs, alpha0, f, df, xt, var_f, var_df] = probLineSearch...
    (ff, xt, f, df, search_direction, alpha0, verbosity, outs, paras, var_f, var_df);

    % -- update search direction ------------------------------------------
    search_direction   = - df;     % new search direction

    % -- storage only for plot --------------------------------------------
    path            = [path, xt];           %#ok<AGROW>
    function_values = [function_values, f]; %#ok<AGROW>
    
end


%% -- plot -----------------------------------------------------------------

clf;

ax = linspace(x1min, x1max, 200);
ay = linspace(x2min, x2max, 200);
[X, Y] = meshgrid(ax, ay); % a x b
XX = [X(:),Y(:)]; % a*b x 2
Z = F(XX);
Z = reshape(Z,size(X));
subplot(1, 1, 1); hold on;
if min(Z(:)) < 0
    Z = Z - min(Z(:));
end

gray_r = 1-gray;
colormap(gray_r)
imagesc([x1min, x1max],[x2min, x2max], log(Z)); axis tight; colorbar;
plot(path(1,:), path(2,:), '-gx')
for i=1:numel(x_min)
    plot(x_min{i}(1), x_min{i}(2), 'ro', 'markersize', 10)
end